"""
Unit test file.
"""

import unittest

from advanced_askai.cli import cli
from advanced_askai.util import authenticaion_valid, get_authentication

IS_AUTHENTICATED = authenticaion_valid()


class InteractiveSessionTester(unittest.TestCase):
    """Main tester class."""

    @unittest.skipUnless(
        IS_AUTHENTICATED, "Authentication required to run this unit test"
    )
    def test_api(self) -> None:
        """Test command line interface (CLI)."""
        # rtn = os.system(COMMAND)
        # self.assertEqual(0, rtn)
        openai_key = get_authentication()
        assert openai_key is not None
        cli()


if __name__ == "__main__":
    unittest.main()



"""
Unit test file.
"""

import io
import sys
import unittest
from unittest.mock import patch

from advanced_askai.cli import cli
from advanced_askai.util import authenticaion_valid

IS_AUTHENTICATED = authenticaion_valid()


class SinglePromptTester(unittest.TestCase):
    """Main tester class."""

    @unittest.skipUnless(
        IS_AUTHENTICATED, "Authentication required to run this unit test"
    )
    def test_single_shot(self) -> None:
        """Test command line interface (CLI)."""
        sys.argv.append("Respond with HELLO")
        try:
            rtn = cli()
        finally:
            sys.argv.pop()
        self.assertEqual(0, rtn)

    @unittest.skipUnless(
        IS_AUTHENTICATED, "Authentication required to run this unit test"
    )
    @patch("sys.stdin", io.StringIO("exit\n"))
    def test_interactive(self) -> None:
        """Test command line interface (CLI)."""
        rtn = cli()
        self.assertEqual(0, rtn)


if __name__ == "__main__":
    unittest.main()



"""
Unit test file.
"""

import os
import unittest

COMMAND = "askai --help"


class MainTester(unittest.TestCase):
    """Main tester class."""

    def test_imports(self) -> None:
        """Test command line interface (CLI)."""
        rtn = os.system(COMMAND)
        self.assertEqual(0, rtn)


if __name__ == "__main__":
    unittest.main()



import argparse
from getpass import getpass

from advanced_askai.chatgpt import is_valid_api_key
from advanced_askai.openaicfg import create_or_load_config, save_config


def load_or_prompt_for_api_key_and_return_config(args: argparse.Namespace) -> dict:
    config = create_or_load_config()
    if args.set_key:
        if is_valid_api_key(args.set_key):
            config["openai_key"] = args.set_key
            save_config(config)
            config = create_or_load_config()
        else:
            raise ValueError("Invalid OpenAI API key provided.")
    elif "openai_key" not in config:
        while True:
            key = getpass(
                "No OpenAI key found, please enter one now (input will be hidden): "
            )
            if is_valid_api_key(key):
                config["openai_key"] = key
                save_config(config)
                break
            else:
                print("Invalid API key. Please try again.")
    else:
        # Check if the existing key is valid
        if not is_valid_api_key(config["openai_key"]):
            print("Existing API key is invalid. Please enter a new one.")
            while True:
                key = getpass("Enter a valid OpenAI API key: ")
                if is_valid_api_key(key):
                    config["openai_key"] = key
                    save_config(config)
                    break
                else:
                    print("Invalid API key. Please try again.")

    return config


def authenticaion_valid() -> bool:
    config = create_or_load_config()
    return "openai_key" in config and is_valid_api_key(config["openai_key"])


def get_authentication() -> str | None:
    config = create_or_load_config()
    key = config.get("openai_key")
    key = key or None
    return key



from advanced_askai.chatgpt import ChatGpt
from advanced_askai.types import ChatBotConfig


def make_chatbot(config: ChatBotConfig) -> ChatGpt:
    return ChatGpt(
        openai_key=config.api_key,
        model=config.model,
        ai_assistant_prompt=config.ai_assistant_prompt,
        max_tokens=config.max_tokens,
    )



"""
Unit test file.
"""

import unittest

from advanced_askai.api import Chat, ChatBotConfig
from advanced_askai.constants import FAST_MODEL
from advanced_askai.util import authenticaion_valid, get_authentication

IS_AUTHENTICATED = authenticaion_valid()


class ApiTester(unittest.TestCase):
    """Main tester class."""

    @unittest.skipUnless(
        IS_AUTHENTICATED, "Authentication required to run this unit test"
    )
    def test_api(self) -> None:
        """Test command line interface (CLI)."""
        # rtn = os.system(COMMAND)
        # self.assertEqual(0, rtn)
        openai_key = get_authentication()
        assert openai_key is not None
        config = ChatBotConfig(
            model=FAST_MODEL,
            api_key=openai_key,
            ai_assistant_prompt="You are a helpful assistant that always returns 'EXPECTED-MESSAGE', and nothing else",
        )
        chat = Chat(config)
        response = chat.query("test-input")
        assert response == "EXPECTED-MESSAGE"


if __name__ == "__main__":
    unittest.main()



"""askai - ask openai for help"""

import argparse
import sys

from advanced_askai.chatgpt import (
    ChatGpt,
    ChatGPTAuthenticationError,
    ChatGPTConnectionError,
    ChatGPTRateLimitError,
    get_max_tokens,
)
from advanced_askai.constants import ADVANCED_MODEL, FAST_MODEL
from advanced_askai.internal_chat_session import (
    internal_interactive_chat_session,
    internal_single_chat_session,
)
from advanced_askai.openaicfg import create_or_load_config, save_config
from advanced_askai.parse_args import parse_args
from advanced_askai.prompt_input import prompt_input
from advanced_askai.streams import ConsoleStream, FileOutputStream, MultiStream, Stream
from advanced_askai.util import load_or_prompt_for_api_key_and_return_config


def get_model_max_tokens(args: argparse.Namespace) -> tuple[str, int]:
    max_tokens = args.max_tokens
    model: str
    if args.model is None:
        if args.advanced:
            model = ADVANCED_MODEL
        else:
            model = FAST_MODEL
    else:
        model = args.model

    if max_tokens is None:
        max_tokens = get_max_tokens(model)
    return model, max_tokens


def cli() -> int:
    args = parse_args()
    config = create_or_load_config()
    if args.input_file:
        with open(args.input_file, "r") as file:
            args.prompt = file.read().strip()
    model: str
    max_tokens: int
    model, max_tokens = get_model_max_tokens(args)

    config = load_or_prompt_for_api_key_and_return_config(args)
    key = config["openai_key"]

    prompt = args.prompt

    def log(*pargs, **kwargs):
        if not args.verbose:
            return
        print(*pargs, **kwargs)

    if args.assistant_prompt_file:
        with open(args.assistant_prompt_file, "r") as f:
            ai_assistant_prompt = f.read().strip()
    else:
        ai_assistant_prompt = args.assistant_prompt

    log(prompt)

    chatbot = ChatGpt(
        openai_key=key,
        max_tokens=max_tokens,
        model=model,
        ai_assistant_prompt=ai_assistant_prompt,
    )

    streams: list[Stream] = [ConsoleStream()]
    if args.output:
        streams.append(FileOutputStream(args.output))
    outstream = MultiStream(streams)

    try:
        interactive = prompt is None
        if not interactive:
            internal_single_chat_session(
                chatbot=chatbot,
                prompts=[prompt],
                outstream=outstream,
                as_json=args.json,
                check=args.check,
                no_stream=args.no_stream,
                status_print_func=print,
            )
            return 0
        print(
            "\nInteractive mode - press return three times to submit your code to OpenAI"
        )
        internal_interactive_chat_session(
            chatbot=chatbot,
            prompts=[],
            outstream=outstream,
            as_json=args.json,
            no_stream=args.no_stream,
            check=args.check,
            prompt_input_func=prompt_input,
            status_print_func=print,
        )
    except ChatGPTConnectionError as err:
        print(err)
        return 1
    except ChatGPTAuthenticationError as e:
        print(
            "Error authenticating with OpenAI, deleting password from config and exiting."
        )
        print(e)
        save_config({})
        return 1
    except ChatGPTRateLimitError:
        print("Rate limit exceeded, set a new key with --set-key")
        return 1
    return 0


def main() -> int:
    try:
        return cli()
    except KeyboardInterrupt:
        return 1
    except SystemExit as e:
        if isinstance(e.code, int):
            return e.code
        return 1


if __name__ == "__main__":
    sys.exit(main())



"""Prompts for input until two empty lines are entered."""

import time


def prompt_input() -> str:
    """Prompts for input until two empty lines are entered."""
    lines: list[str] = []
    empty_count = 0
    response_times: list[float] = []
    while True:
        start_time = time.time()
        line = input(">>> ")
        if not lines and (line.startswith("!") or line == "exit"):
            return "\n".join(lines + [line])
        diff_time = time.time() - start_time
        response_times.append(diff_time)
        if not line:
            empty_count += 1
            if empty_count == 2:
                last_two = response_times[-2:]
                avg = sum(last_two) / len(last_two)
                if avg < 0.1:
                    empty_count = 0
                else:
                    break
        else:
            empty_count = 0
        lines.append(line)
    return "\n".join(lines)



"""
Unit test file.
"""

import unittest
from typing import Callable

from advanced_askai.chatgpt import ChatGpt
from advanced_askai.constants import FAST_MODEL
from advanced_askai.internal_chat_session import internal_interactive_chat_session
from advanced_askai.streams import ConsoleStream
from advanced_askai.util import authenticaion_valid, get_authentication

IS_AUTHENTICATED = authenticaion_valid()


def _fake_prompt_generator(return_values: list[str]) -> Callable[[], str]:
    idx = 0

    def func() -> str:
        nonlocal return_values
        nonlocal idx
        out = return_values[idx % len(return_values)]
        idx += 1
        assert idx <= len(return_values)
        return out

    return func


class InteractiveSessionTester(unittest.TestCase):
    """Main tester class."""

    @unittest.skipUnless(
        IS_AUTHENTICATED, "Authentication required to run this unit test"
    )
    def test_api(self) -> None:
        """Test command line interface (CLI)."""
        # rtn = os.system(COMMAND)
        # self.assertEqual(0, rtn)
        openai_key = get_authentication()
        assert openai_key is not None

        chatbot = ChatGpt(
            openai_key=openai_key,
            max_tokens=1024,
            model=FAST_MODEL,
            ai_assistant_prompt="You are a helpful assistant that always returns 'EXPECTED-MESSAGE', and nothing else",
        )
        internal_interactive_chat_session(
            chatbot=chatbot,
            prompts=[],
            outstream=ConsoleStream(),
            as_json=False,
            no_stream=True,
            check=False,
            prompt_input_func=_fake_prompt_generator(
                ["just respond with 'ok'", "just respond with 'no'", "exit"]
            ),
            status_print_func=print,
        )


if __name__ == "__main__":
    unittest.main()



"""askai - ask openai for help"""

from advanced_askai.chatgpt import ChatGpt, ChatStream
from advanced_askai.streams import Stream


def run_chat_query(
    chatbot: ChatGpt,
    prompts: list[str],
    outstream: Stream,
    as_json: bool,
    no_stream: bool,
    print_status: bool,
) -> str:
    """Runs a chat query, throws exceptions if there are issues. Returns the response text."""
    if not as_json:
        if print_status:
            print("############ OPEN-AI QUERY")
    chat_stream: ChatStream = chatbot.query(prompts, no_stream=no_stream)
    if as_json:
        return str(chat_stream)

    if chat_stream is None or not chat_stream.success():
        if print_status:
            print("No error response received from OpenAI, response was:")
        outstream.write(str(chat_stream.response()))
        raise Exception("No response from OpenAI")

    if print_status:
        print("############ OPEN-AI RESPONSE\n")
    response_text = ""
    for text in chat_stream:
        if text is None:
            break
        response_text += text
        outstream.write(text)
    outstream.write("\n\n")

    # return None
    return response_text



import os
from typing import Callable

from advanced_askai.chatgpt import ChatGpt
from advanced_askai.constants import AI_ASSISTANT_CHECKER_PROMPT
from advanced_askai.run_chat_query import run_chat_query
from advanced_askai.streams import NullOutStream, Stream


def internal_iterate_chat_session(
    chatbot: ChatGpt,
    prompts: list[str],
    outstream: Stream,
    as_json: bool,
    no_stream: bool,
    check: bool,
    status_print_func: Callable[[str], None],
) -> str:
    """Runs a single chat query, throws exceptions if there are issues"""
    assert len(prompts) % 2 == 1, "Prompt count should be odd"
    response_text = run_chat_query(
        chatbot=chatbot,
        prompts=prompts,
        outstream=NullOutStream() if check else outstream,
        as_json=as_json,
        no_stream=no_stream,
        print_status=True,
    )
    if not check:
        return response_text
    prompts.append(response_text)
    check_prompt = AI_ASSISTANT_CHECKER_PROMPT + "\n\n" + prompts[-1]
    prompts.append(check_prompt)
    status_print_func("\n############ CHECKING RESPONSE")
    response_text = run_chat_query(
        chatbot=chatbot,
        prompts=prompts,
        outstream=outstream,
        as_json=as_json,
        no_stream=no_stream,
        print_status=False,
    )
    return response_text


def internal_single_chat_session(
    chatbot: ChatGpt,
    prompts: list[str],
    outstream: Stream,
    as_json: bool,
    no_stream: bool,
    check: bool,
    status_print_func: Callable[[str], None],
) -> str:
    """Runs a single chat query, throws exceptions if there are issues"""
    return internal_iterate_chat_session(
        chatbot=chatbot,
        prompts=prompts,
        outstream=outstream,
        as_json=as_json,
        no_stream=no_stream,
        check=check,
        status_print_func=status_print_func,
    )


def internal_interactive_chat_session(
    chatbot: ChatGpt,
    prompts: list[str],
    outstream: Stream,
    as_json: bool,
    no_stream: bool,
    check: bool,
    prompt_input_func: Callable[[], str],
    status_print_func: Callable[[str], None],
) -> None:
    """Runs a chat query, throws exceptions if there are issues"""
    assert (
        len(prompts) % 2 == 0
    ), "Pre-populated prompts (indicating previous history) should be even"

    while True:
        next_prompt = prompt_input_func()
        prompts.append(next_prompt)
        new_cmd = prompts[-1].strip().replace("()", "")
        if new_cmd.startswith("!"):
            prompts = prompts[0:-1]
            new_cmd = new_cmd[1:]
            rtn = os.system(new_cmd)
            status_print_func(f"Command exited and returned {rtn}")
            next_prompt = prompt_input_func()
            prompts.append(next_prompt)
            continue
        elif new_cmd == "exit":
            status_print_func("Exited due to 'exit' command")
            return
        assert len(prompts) % 2 == 1, "Prompt count should be odd"
        response_text = internal_iterate_chat_session(
            chatbot=chatbot,
            prompts=prompts,
            outstream=outstream,
            as_json=as_json,
            no_stream=no_stream,
            check=check,
            status_print_func=status_print_func,
        )
        prompts.append(response_text)



import argparse

from advanced_askai.constants import ADVANCED_MODEL, AI_ASSISTANT


def parse_args() -> argparse.Namespace:
    argparser = argparse.ArgumentParser(usage="Ask OpenAI for help with code")
    argparser.add_argument("prompt", help="Prompt to ask OpenAI", nargs="?")
    argparser.add_argument(
        "--input-file", help="Input file containing prompts", type=str
    )
    argparser.add_argument("--json", help="Print response as json", action="store_true")
    argparser.add_argument("--set-key", help="Set OpenAI key")
    argparser.add_argument("--output", help="Output file", type=str)
    model_group = argparser.add_mutually_exclusive_group()
    model_group.add_argument(
        "--advanced",
        action="store_true",
        default=False,
        help=f"bleeding edge model: {ADVANCED_MODEL}",
    )
    model_group.add_argument("--model", default=None)
    argparser.add_argument("--verbose", action="store_true", default=False)
    argparser.add_argument("--no-stream", action="store_true", default=False)
    argparser.add_argument("--assistant-prompt", type=str, default=AI_ASSISTANT)
    argparser.add_argument(
        "--assistant-prompt-file", type=str, help="File containing assistant prompt"
    )
    # max tokens
    argparser.add_argument(
        "--max-tokens", help="Max tokens to return", type=int, default=None
    )
    argparser.add_argument(
        "--code",
        action="store_true",
        default=False,
        help="Code mode: enables aider mode",
    )
    argparser.add_argument(
        "--check",
        action="store_true",
        default=False,
        help="Sends the response back to the chatbot for a second opinion",
    )
    return argparser.parse_args()



"""
Unit test file.
"""

import subprocess
import unittest

from advanced_askai.util import authenticaion_valid

IS_AUTHENTICATED = authenticaion_valid()


class SingleCheckedPromptTester(unittest.TestCase):
    """Main tester class."""

    @unittest.skipUnless(
        IS_AUTHENTICATED, "Authentication required to run this unit test"
    )
    def test_cli(self) -> None:
        command = 'askai "Respond with HELLO" --check'
        output = subprocess.check_output(command, shell=True, text=True)
        self.assertIn("HELLO", output)

    @unittest.skipUnless(
        IS_AUTHENTICATED, "Authentication required to run this unit test"
    )
    def test_cli_advanced(self) -> None:
        command = 'askai "Respond with HELLO" --advanced --check'
        output = subprocess.check_output(command, shell=True, text=True)
        self.assertIn("HELLO", output)


if __name__ == "__main__":
    unittest.main()



# mypy: disable-error-code="attr-defined,valid-type"

"""Interacts with open ai's chat bot."""


import threading
from typing import Any, Optional

import json5 as json
import openai
import tiktoken
from openai import APIConnectionError, AuthenticationError, BadRequestError, OpenAI

from advanced_askai.constants import ADVANCED_MODEL, HIDDEN_PROMPT_TOKEN_COUNT


def is_valid_api_key(api_key: str) -> bool:
    """
    Test whether the given OpenAI API key is valid.

    :param api_key: The OpenAI API key to test
    :return: True if the key is valid, False otherwise
    """
    client = OpenAI(api_key=api_key)
    try:
        # Attempt to list models, which requires a valid API key
        client.models.list()
        return True
    except AuthenticationError:
        return False
    except (APIConnectionError, BadRequestError):
        # If there's a connection error or bad request, we can't determine if the key is valid
        raise


def get_max_tokens(model: str) -> int:
    if model == ADVANCED_MODEL:
        return 4096
    return 16384


class ChatGPTConnectionError(BaseException):
    def __init__(self, conn_error: APIConnectionError):
        super().__init__(conn_error)
        self.err = conn_error


class ChatGPTAuthenticationError(BaseException):
    def __init__(self, auth_error: AuthenticationError):
        super().__init__(auth_error)
        self.err = auth_error


class ChatGPTRateLimitError(BaseException):
    def __init__(self, rate_error: openai.RateLimitError):
        super().__init__(rate_error)
        self.err = rate_error


ChatCompletion = openai.ChatCompletion  # pylint: disable=no-member


# Create a thread-safe dictionary to store OpenAI client instances
client_instances_lock = threading.Lock()
client_instances: dict[str, openai.OpenAI] = {}


def get_client_instance(openai_key: str) -> OpenAI:
    with client_instances_lock:
        if openai_key not in client_instances:
            client_instances[openai_key] = openai.OpenAI(api_key=openai_key)
        out = client_instances[openai_key]
        assert out is not None
        return out


def count_tokens(model: str, text: str):
    enc = tiktoken.encoding_for_model(model)
    return len(enc.encode(text))


def ai_query(
    openai_key: str,
    prompts: list[str],
    max_tokens: int,
    model: str,
    ai_assistant_prompt: str,
) -> ChatCompletion:
    # assert prompts is odd
    assert (
        len(prompts) % 2 == 1
    )  # Prompts alternate between user message and last response
    messages = [
        {
            "role": "system",
            "content": ai_assistant_prompt,
        },
    ]
    for i, prompt in enumerate(prompts):
        if i % 2 == 0:
            messages.append({"role": "assistant", "content": prompt})
        else:
            messages.append({"role": "user", "content": prompt})

    # Use the get_client_instance function to ensure thread safety
    client = get_client_instance(openai_key)

    # compute the max_tokens by counting the tokens in the prompt
    # and subtracting that from the max_tokens
    messages_json_str = json.dumps(messages)
    prompt_tokens = count_tokens(model, messages_json_str)
    max_tokens = max_tokens - prompt_tokens - HIDDEN_PROMPT_TOKEN_COUNT
    try:
        response = client.chat.completions.create(
            model=model,
            messages=messages,  # type: ignore
            temperature=0.7,
            max_tokens=max_tokens,
            top_p=0.3,
            frequency_penalty=0.5,
            presence_penalty=0,
            stream=True,
        )
        return response
    except APIConnectionError as e:
        raise ChatGPTConnectionError(e)
    except AuthenticationError as e:
        raise ChatGPTAuthenticationError(e)


class ChatStream:
    def __init__(self, chatcompletion: ChatCompletion, no_stream: bool = False):
        self.chatcompletion = chatcompletion
        self.no_stream = no_stream

    def success(self) -> bool:
        return self.chatcompletion.response.is_success

    def response(self) -> Any:
        return self.chatcompletion.response

    def __iter__(self):
        all_str = ""
        for event in self.chatcompletion:
            choice = event.choices[0]
            delta = choice.delta
            event_text = delta.content
            if event_text is None:
                break
            if not self.no_stream:
                yield event_text
            else:
                all_str += event_text
        if self.no_stream:
            yield all_str


class ChatGpt:
    def __init__(
        self,
        openai_key: str,
        model: str,
        ai_assistant_prompt: str,
        max_tokens: int | None = None,
    ):
        self.openai_key = openai_key
        self.model = model
        self.max_tokens = max_tokens or get_max_tokens(model)
        self.ai_assistant_prompt = ai_assistant_prompt

    def query(
        self,
        prompts: list[str],
        max_tokens: Optional[int] = None,
        no_stream: bool = False,
    ) -> ChatStream:
        chat_completion = ai_query(
            openai_key=self.openai_key,
            prompts=prompts,
            max_tokens=max_tokens or self.max_tokens,
            model=self.model,
            ai_assistant_prompt=self.ai_assistant_prompt,
        )
        return ChatStream(chat_completion, no_stream)



from typing import Callable

from advanced_askai.internal_chat_session import (
    internal_interactive_chat_session,
    internal_single_chat_session,
)
from advanced_askai.make_chatbot import make_chatbot
from advanced_askai.prompt_input import prompt_input
from advanced_askai.streams import ConsoleStream, NullOutStream, Stream
from advanced_askai.types import ChatBotConfig


def chat_session(
    chat_config: ChatBotConfig,
    prompts: list[str] | None = None,
    outstream: Stream | None = None,
    as_json: bool = False,
    no_stream: bool = False,
    check: bool = False,
    status_print_func: Callable[[str], None] = print,
) -> None:
    """Runs a single chat query, throws exceptions if there are issues"""
    prompts = prompts or []
    outstream = outstream or ConsoleStream()
    chatbot = make_chatbot(chat_config)
    return internal_interactive_chat_session(
        chatbot=chatbot,
        prompts=prompts,
        outstream=outstream,
        as_json=as_json,
        no_stream=no_stream,
        check=check,
        status_print_func=status_print_func,
        prompt_input_func=prompt_input,
    )


def single_query(
    chat_config: ChatBotConfig,
    prompt: str | list[str],
    outstream: Stream | None = None,
    as_json: bool = False,
    no_stream: bool = False,
    check: bool = False,
    status_print_func: Callable[[str], None] = print,
) -> str:
    """Runs a single chat query, throws exceptions if there are issues"""
    chatbot = make_chatbot(chat_config)
    outstream = outstream or ConsoleStream()
    if isinstance(prompt, str):
        prompts = [prompt]
    else:
        prompts = prompt
    return internal_single_chat_session(
        chatbot=chatbot,
        prompts=prompts,
        outstream=outstream,
        as_json=as_json,
        no_stream=no_stream,
        check=check,
        status_print_func=status_print_func,
    )


class Chat:
    def __init__(
        self,
        chat_config: ChatBotConfig,
        chat_history: list[str] | None = None,
        outstream: Stream | None = None,
        as_json: bool = False,
        no_stream: bool = False,
        check: bool = False,
        status_print_func: Callable[[str], None] = print,
    ) -> None:
        self.chat_config = chat_config
        self.prompts = chat_history if chat_history is not None else []
        self.outstream = outstream or NullOutStream()
        self.as_json = as_json
        self.no_stream = no_stream
        self.check = check
        self.status_print_func = status_print_func

    def query(self, prompt: str) -> str:
        self.prompts.append(prompt)
        response_text = single_query(
            chat_config=self.chat_config,
            prompt=prompt,
            outstream=self.outstream,
            as_json=self.as_json,
            no_stream=self.no_stream,
            check=self.check,
            status_print_func=self.status_print_func,
        )
        self.prompts.append(response_text)
        return response_text



import sys
from abc import ABC, abstractmethod
from typing import List


class Stream(ABC):
    @abstractmethod
    def write(self, text: str) -> None:
        pass

    def close(self) -> None:
        pass


class FileOutputStream(Stream):
    def __init__(self, outfile: str):
        self.outfile = outfile

    def write(self, text: str) -> None:
        with open(self.outfile, "a") as f:
            f.write(text)


class ConsoleStream(Stream):
    def __init__(self):
        pass

    def write(self, text: str) -> None:
        sys.stdout.write(text)
        sys.stdout.flush()


class MultiStream(Stream):
    def __init__(self, streams: List[Stream]):
        self.streams = streams

    def write(self, text: str) -> None:
        for stream in self.streams:
            stream.write(text)

    def close(self) -> None:
        for stream in self.streams:
            stream.close()


class NullOutStream(Stream):
    def write(self, text: str) -> None:
        pass



HIDDEN_PROMPT_TOKEN_COUNT = (
    100  # this hack corrects for the unnaccounted for tokens in the prompt
)
ADVANCED_MODEL = "gpt-4o"
FAST_MODEL = "gpt-4o-mini"

AI_ASSISTANT = (
    "You are a helpful assistant to a senior programmer. "
    "If I am asking how to do something in general then go ahead "
    "and recommend popular third-party apps that can get the job done, "
    "but don't recommend additional tools when I'm currently asking how to do use "
    "a specific tool."
)

AI_ASSISTANT_CHECKER_PROMPT = (
    "Now check this answer and verify if it's correct. You will look at the question asked and\n"
    "check that the answer matches the exact specifications. If it's correct, say so. \n"
    "If it's not correct or if there are any issues, explain what's wrong and provide the correct information. \n"
    "Otherwise, say that this correct and repeat the answer VERBATIM.\n"
    'Do not provide any additional information, such as "This is correct." or "This is incorrect.", just provide the correct information.'
)



import os
import re
import shutil
import sys
from typing import Optional

import colorama
from colorama import AnsiToWin32
from consolemd import Renderer as ConsoleMDRenderer

colorama.init(wrap=False)

_LINE_UP = "\033[1A"
_LINE_CLEAR = "\x1b[2K"
_LINE_LEFT = "\r"
_term_stdout = AnsiToWin32(sys.stdout, autoreset=True).stream
ansi_escape_regex = re.compile(r"\x1b\[[0-9;]*[a-zA-Z]")


def _num_terminal_cols() -> int:
    cols, _ = shutil.get_terminal_size()
    return cols


def get_options() -> dict:
    options = {
        "end-of-line": "keep",
    }
    return options


def _get_clear_line_sequence(n: int) -> str:
    out = ""
    for _ in range(n):
        out += _LINE_UP + _LINE_CLEAR + _LINE_LEFT
    return out


def count_displaying_characters2(text):
    in_escape_sequence = False
    displaying_char_count = 0

    for char in text:
        if char == "\x1b":  # Start of an ANSI escape sequence
            in_escape_sequence = True
        elif in_escape_sequence and char.isalpha():  # End of an ANSI escape sequence
            in_escape_sequence = False
        elif not in_escape_sequence:  # Counting characters outside of ANSI sequences
            displaying_char_count += 1

    return displaying_char_count


def _count_overflow_lines(text: str, width: int) -> int:
    lines = text.split("\n")
    num_lines = 0
    for line in lines:
        num_display_chars = len(line)
        if num_display_chars > width:
            num_lines += num_display_chars // width
    return num_lines


class MyStream:
    def __init__(self):
        self.buffer = ""

    def write(self, text: str) -> None:
        self.buffer += text

    def flush(self) -> None:
        pass

    def close(self) -> None:
        pass

    def read(self) -> str:
        return self.buffer


def _to_consolemd2(text: str) -> str:
    stringio = MyStream()
    style = os.environ.get("CONSOLEMD_STYLE", "native")
    renderer = ConsoleMDRenderer(style_name=style)
    kwargs = {
        "output": stringio,
        "soft_wrap": False,
    }
    renderer.render(text, **kwargs)
    out = stringio.read()
    return out


def _color_print(text: str, delete_lines: Optional[int] = 0) -> str:
    stdout = _to_consolemd2(text=text)
    if delete_lines:
        clear_msg = _get_clear_line_sequence(delete_lines)
        stdout = clear_msg + stdout
    _term_stdout.write(stdout)
    sys.stdout.flush()
    return stdout


class StreamingConsoleMarkdown:
    def __init__(self):
        self.last_written_lines = 0
        self.cols = None
        self.last_written_md = ""

    def update(self, text: str) -> str:
        text = "\n" + text
        cols = self.cols or _num_terminal_cols()
        delete_lines = self.last_written_lines

        # self.last_written_lines = _count_text_lines(text, cols)
        self.last_written_md = _color_print(text, delete_lines=delete_lines)
        lines = self.last_written_md.split("\n")
        n_lines = len(lines) - 1
        n_overflows = _count_overflow_lines(text, cols)
        self.last_written_lines = n_lines + n_overflows
        return self.last_written_md

    def pop_last(self) -> None:
        if self.last_written_lines:
            clear_msg = _get_clear_line_sequence(self.last_written_lines + 1)
            _term_stdout.write("\n" + clear_msg)
            sys.stdout.flush()
            self.last_written_lines = 0
            self.last_written_md = ""

    def _clear_last_line(self) -> None:
        clr_msg = _get_clear_line_sequence(2)
        sys.stdout.write("\n" + clr_msg)
        sys.stdout.flush()



import json
import os

import keyring  # type: ignore

# from appdirs import user_config_dir  # type: ignore

SERVICE_NAME = "advanced-askai"
# get the current user name
USERNAME = os.getlogin()


def _set_data(data: str) -> None:
    keyring.set_password(SERVICE_NAME, USERNAME, data)


def _get_data() -> str | None:
    return keyring.get_password(SERVICE_NAME, USERNAME)


def save_config(config: dict) -> None:
    data_str = json.dumps(config)
    _set_data(data_str)


def create_or_load_config() -> dict:
    try:
        data = _get_data()
        if data is not None:
            return json.loads(data)
        save_config({})
        return {}
    except OSError:
        save_config({})
        return {}
